import functools
import itertools

import pydash.arrays as arrays
import pydash.collections as collections


find = collections.find

def count(gen):
    return functools.reduce(lambda x,y: x + 1, gen, 0)


def flatten(gen):
    for arr in gen:
        for elem in arrays.flatten(arr, is_deep=True):
            yield elem


def split(gen, n=1):
    """
    Yield successive n-sized chunks from gen.
    Sample usage:
        a = (_ for _ in range(28):
        for chunked in tdash.split(a, n=10):
            print list(chunked)
    Expected output:
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        [10, 11, 12, 13, 14, 15, 16, 17, 18, 19]
        [20, 21, 22, 23, 24, 25, 26, 27]

    :param gen
    :param n
    """
    it = iter(gen)
    for first in it:
        yield itertools.chain([first], itertools.islice(it, n - 1))

# ALIAS
chunk = split


import operator
import functools
import warnings

class LRUCache(object):
    def __init__(self, size, provider_func, on_evicted_func=None, on_entry_func=None):
        """
        :param size: size of cache
        :type size: int
        :param provider_func: gets the value if missing from cache of will be called with the regular key
                              ie func(key) and should return value.
        :type provider_func: func if None then do nothing
        :param on_evicted_func: gets called on the key, value when being removed from the cache.
                               ie func(key, value). whatever it returns is not used.
        :type on_evicted_func: func
        :param on_entry_func: gets called when function is entered into the cache via func(key, value).
                              if None does nothing
        :type on_entry_func: func
        """
        self._size = size
        self._counter = 0
        self._provider_func = provider_func
        self._on_evicted_func = functools.partial(call_func_if_not_none, on_evicted_func)
        self._on_entry_func = functools.partial(call_func_if_not_none, on_entry_func)

        self._values = {}
        self._access_history = {}

    def __getitem__(self, key):
        self._note_access(key)
        try:
            return self._values[key]
        except KeyError:
            return self._populate_missing(key)

    def __setitem__(self, key, value):
        self._on_entry_func(key, value)
        self._note_access(key)
        self._values[key] = value
        self._evict_lru_if_applicable()

    def __len__(self):
        return len(self._values)

    def _populate_missing(self, key):
        self._values[key] = self._provider_func(key)
        self._evict_lru_if_applicable()
        return self._values[key]

    def _evict_lru_if_applicable(self):
        if len(self) > self._size:
            lru_key, _ = min(self._access_history.items(), key=operator.itemgetter(1))
            failed = False
            # TODO: check if two equal sized item what happens.
            # TODO synchronise this function to stop multiple threads.
            #TODO (cont): see https://github.com/GrahamDumpleton/wrapt/blob/develop/blog/07-the-missing-synchronized-decorator.md

            try:
                lru_val = self._values.pop(lru_key)
            except KeyError:
                failed = True
                warnings.warn('Cannot find key %s in values. Values keys are \n %s. \n Access History items are \n %s' % (
                    str(lru_key), str(self._values.keys()), str(self._access_history.items())))

            try:
                del self._access_history[lru_key]
            except KeyError:
                warnings.warn(
                    'Cannot find key %s in access history. Values keys are \n %s. \n Access History items are \n %s' % (
                        str(lru_key), str(self._values.keys()), str(self._access_history.items())))

            # Do this last if possible as may take some time.
            if not failed:
                self._on_evicted_func(lru_key, lru_val)

    def _note_access(self, key):
        self._access_history[key] = self._get_counter_and_increase_by_one()

    def _get_counter_and_increase_by_one(self):
        old_counter = self._counter
        self._counter += 1
        return old_counter


def call_func_if_not_none(func, *args, **kwargs):
    if func is not None:
        return func(*args, **kwargs)





from functools import wraps
from time import time


def measure_time(name=None):
    def decorator(func):
        called = name if name else func.__name__

        @wraps(func)
        def func_wrapper(*args, **kwargs):
            t0 = time()
            result = func(*args, **kwargs)
            t = time() - t0
            print(called, t)
            return result
        return func_wrapper
    return decorator


class Stopwatch(object):
    def __init__(self, inside_mg='Default watch,'):
        self.start_time = time()
        self.inside_mg = inside_mg

    def print_time(self, msg):
        print(self.inside_mg, 'Time: ', msg, ':   ', time() - self.start_time)
        self.start_time = time()



import time
import functools
import inspect
from threading import RLock

from .generators import count


def retry_n_times(fn, n, exception=Exception, interval=0, on_exception=None, args=(), kwargs=None):
    if kwargs is None:
        kwargs = {}  # dictionarys are mutable so not ideal for default parameter values.

    for i in range(n):
        if i == n-1:  # Last try so don't catch exception
            return fn(*args, **kwargs)
        try:
            return fn(*args, **kwargs)
        except exception as e:
            if interval > 0:
                time.sleep(interval)
            if on_exception:
                on_exception(e)


###
# Function composition is a way of combining functions such that the result of each function
# is passed as the argument of the next function.
# For example, the composition of two functions f and g is f(g(x)) or compose(f,g)(x)
###
def compose(*funcs): return lambda x: reduce(lambda v, f: f(v), reversed(funcs), x)

###
# Sometimes when writing functional code it's handy to raise an exception
# there doesnt seem to be a built in function that raises an exception (raise is a statememt)
# e.g. you cannot write lambda x: raise Exception('bugger')
# you can howerver write lambda x: _.raise(Exception('bugger'))
###
def raise_ex(exception):
    raise exception

###
# Like partial but it safely ignores excess arguments and keywords!
###
def safe_partial(fn, *args, **kwargs):

    fn_args, fn_varargs, fn_keywords, __ = inspect.getargspec(fn)

    if fn_keywords is None:
        kwargs = { k: v for k,v in kwargs.items() if k in fn_args }

    available_arguments =  len(fn_args) - count((k for k in kwargs.keys() if k in fn_args ))
    if fn_varargs is None and len(args) > available_arguments:
        # Throw away some arguments!
        args = args[0:len(fn_args)]

    return functools.partial(fn, *args, **kwargs)

###
# Convenience method for wrapping a function on the fly.
# Wrapper function takes a function, which is the thing it is wrapping with its args already curried in as first argument
# you cannot mess therefore with the arguments. But you can decide to execute it or not and do stuff before or after
###
def wrap(fn, wrapper):
    @functools.wraps(fn)
    def wrap(*args, **kwds):
        next = functools.partial(fn, *args, **kwds)
        wrapper_args = (next,) + args
        return safe_partial(wrapper, *wrapper_args, **kwds)()
    return wrap


###
# Method for exception handling on the fly.
# n.b. executes f
# n.b. tightly bound to on_exception
###
def attempt(fn, on_exception, catch_exception=Exception):
    try:
        return fn()
    except catch_exception as e:
        return on_exception(e)

###
# Convenience method.
# Returns a function wrapped so that exceptions are handled and caught with default method
# default method receives the thrown exception as arg
###
def on_exception(fn, on_exception, catch_exception=Exception):
    return wrap(fn, lambda f: attempt(f, on_exception=on_exception, catch_exception=catch_exception))


def result(fn_or_value, *args, **kwargs):
    """
     If argument is a function execute it. If it is a value, return it.
     Useful for handling lazy execution of default cases:
     e.g.
     _.result(some_dict.get(key, lambda: ))
    :param fn_or_value:
    :param args:
    :param kwargs:
    :return:
    """
    if inspect.isfunction(fn_or_value):
        return safe_partial(fn_or_value, *args, **kwargs)()
    return fn_or_value


class cache_function(object):
    def __init__(self, key_fn=lambda x:x):
        self.key_fn = key_fn
        self.cache = dict()
        self.lock = RLock()

    def __call__(self, f):
        def decorated_function(*args):
            key = self.key_fn(*args)
            self.lock.acquire()
            try:
                if key in self.cache:
                    return self.cache[key]
                else:
                    self.cache[key] = f(*args)
                    return self.cache[key]
            finally:
                self.lock.release()
        setattr(decorated_function,'_cache', self.cache)
        return decorated_function

def memoize(f):
    """
    Decorator to memoize a function taking one or more arguments.
    Uses per-instance cache. Does not have a method to reset cache.
    :param f: Function to be decorated
    : return:
    """
    cache = {}

    def decorated_function(*args):
        key = args
        if key in cache:
            return cache[key]
        else:
            cache[key] = f(*args)
            return cache[key]
    return decorated_function
